import httpx

from kleinkram.auth.auth import TokenFile, CLI_KEY, AUTH_TOKEN, REFRESH_TOKEN


class NotAuthenticatedException(Exception):

    def __init__(self, endpoint: str):
        self.message = f"You are not authenticated on endpoint '{endpoint}'. Please run 'klein login' to authenticate."


class AuthenticatedClient(httpx.Client):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        try:
            self.tokenfile = TokenFile()
            self._load_cookies()
        except Exception:
            raise NotAuthenticatedException(self.tokenfile.endpoint)

    def _load_cookies(self):
        if self.tokenfile.isCliToken():
            self.cookies.set(CLI_KEY, self.tokenfile.getCLIToken())
        else:
            self.cookies.set(AUTH_TOKEN, self.tokenfile.getAuthToken())

    def refresh_token(self):
        if self.tokenfile.isCliToken():
            print("CLI key cannot be refreshed.")
            return
        refresh_token = self.tokenfile.getRefreshToken()
        if not refresh_token:
            print("No refresh token found. Please login again.")
            raise Exception("No refresh token found.")
        self.cookies.set(
            REFRESH_TOKEN,
            refresh_token,
        )
        response = self.post(
            "/auth/refresh-token",
        )
        response.raise_for_status()
        new_access_token = response.cookies.get(AUTH_TOKEN)
        new_tokens = {AUTH_TOKEN: new_access_token, REFRESH_TOKEN: refresh_token}
        self.tokenfile.saveTokens(new_tokens)
        self.cookies.set(AUTH_TOKEN, new_access_token)

    def request(self, method, url, *args, **kwargs):
        response = super().request(
            method, self.tokenfile.endpoint + url, *args, **kwargs
        )
        if (url == "/auth/refresh-token") and response.status_code == 401:
            print("Refresh token expired. Please login again.")
            response.status_code = 403
            exit(1)
        if response.status_code == 401:
            print("Token expired, refreshing token...")
            self.refresh_token()
            response = super().request(
                method, self.tokenfile.endpoint + url, *args, **kwargs
            )
        return response



import json
import os
import urllib.parse
import webbrowser
from http.server import BaseHTTPRequestHandler, HTTPServer
from pathlib import Path
from typing import Optional

import typer
from typing_extensions import Annotated

from kleinkram.consts import API_URL

TOKEN_FILE = Path(os.path.expanduser("~/.kleinkram.json"))
REFRESH_TOKEN = "refreshtoken"
AUTH_TOKEN = "authtoken"
CLI_KEY = "clikey"


class TokenFile:
    def __init__(self):
        try:
            with TOKEN_FILE.open("r") as token_file:
                content = json.load(token_file)
                self.endpoint = content["endpoint"]
                self.tokens = content["tokens"]
        except FileNotFoundError:
            self.tokens = {}
            self.endpoint = API_URL
        except json.JSONDecodeError:
            print("Token file is corrupted. Please run 'login' command again.")
            raise

    def isCliToken(self):
        return CLI_KEY in self.tokens[self.endpoint]

    def getAuthToken(self):
        return self.tokens[self.endpoint][AUTH_TOKEN]

    def getRefreshToken(self):
        return self.tokens[self.endpoint][REFRESH_TOKEN]

    def getCLIToken(self):
        return self.tokens[self.endpoint][CLI_KEY]

    def writeToFile(self):
        res = {
            "endpoint": self.endpoint,
            "tokens": self.tokens,
        }
        with TOKEN_FILE.open("w") as token_file:
            json.dump(res, token_file)

    def saveTokens(self, tokens):
        self.tokens[self.endpoint] = tokens
        self.writeToFile()


class OAuthCallbackHandler(BaseHTTPRequestHandler):
    def do_GET(self):
        if self.path.startswith("/cli/callback"):
            query = urllib.parse.urlparse(self.path).query
            params = urllib.parse.parse_qs(query)
            self.server.tokens = {
                AUTH_TOKEN: params.get(AUTH_TOKEN)[0],
                REFRESH_TOKEN: params.get(REFRESH_TOKEN)[0],
            }
            self.send_response(200)
            self.send_header("Content-type", "text/html")
            self.end_headers()
            self.wfile.write(b"Authentication successful. You can close this window.")
            return
        print("here")

    def log_message(self, format, *args):
        pass


def get_auth_tokens():
    server_address = ("", 8000)
    httpd = HTTPServer(server_address, OAuthCallbackHandler)
    httpd.handle_request()
    return httpd.tokens


def logout():
    """
    Logout from the currently set endpoint.
    """
    tokenfile = TokenFile()
    tokenfile.tokens[tokenfile.endpoint] = {}
    tokenfile.writeToFile()
    print("Logged out.")


def login(
    key: Optional[str] = typer.Option(None, help="CLI Key", hidden=True),
    open_browser: Optional[bool] = typer.Option(
        True, help="Open browser for authentication"
    ),
):
    """
    Login into the currently set endpoint.\n
    By default, it will open the browser for authentication. On machines without a browser, you can manually open the URL provided and paste the tokens back.
    """
    tokenfile = TokenFile()
    if key:
        token = {}
        token[CLI_KEY] = key
        tokenfile.saveTokens(token)

    else:
        url = tokenfile.endpoint + "/auth/google?state=cli"

        has_browser = True
        try:
            browser_available = webbrowser.get()
            if not browser_available:
                raise Exception("No web browser available.")
        except Exception as e:
            has_browser = False

        if has_browser and open_browser:
            webbrowser.open(url)
            auth_tokens = get_auth_tokens()

            if not auth_tokens:
                raise Exception("Failed to get authentication tokens.")

            tokenfile.saveTokens(auth_tokens)
            print("Authentication complete. Tokens saved to ~/.kleinkram.json.")
            return

        print(
            f"Please open the following URL manually in your browser to authenticate: {url + '-no-redirect'}"
        )
        print("Enter the authentication token provided after logging in:")
        manual_auth_token = input("Authentication Token: ")
        manual_refresh_token = input("Refresh Token: ")
        if manual_auth_token:
            tokenfile.saveTokens(
                {AUTH_TOKEN: manual_auth_token, REFRESH_TOKEN: manual_refresh_token}
            )
            print("Authentication complete. Tokens saved to tokens.json.")
        else:
            raise ValueError("Failed to get authentication tokens.")


def setCliKey(key: Annotated[str, typer.Argument(help="CLI Key")]):
    """
    Set the CLI key (Actions Only)

    Same as login with the --key option.
    Should never be used by the user, only in docker containers launched from within Kleinkram.
    """
    tokenfile = TokenFile()
    if not tokenfile.endpoint in tokenfile.tokens:
        tokenfile.tokens[tokenfile.endpoint] = {}
    tokenfile.tokens[tokenfile.endpoint][CLI_KEY] = key
    tokenfile.writeToFile()



from typing import Annotated

import httpx
import typer
from rich.console import Console
from rich.table import Table

from kleinkram.api_client import AuthenticatedClient

tag = typer.Typer(
    name="tag",
    help="Tag operations",
    no_args_is_help=True,
    context_settings={"help_option_names": ["-h", "--help"]},
)


@tag.command("list-tag-types")
def tag_types(
    verbose: Annotated[bool, typer.Option()] = False,
):
    """
    List all tagtypes
    """

    try:
        client = AuthenticatedClient()
        response = client.get("/tag/all")
        response.raise_for_status()
        data = response.json()

        if not data or len(data) == 0:
            print("No tagtypes found")
            return

        if verbose:
            table = Table("UUID", "Name", "Datatype")
            for tagtype in data:
                table.add_row(tagtype["uuid"], tagtype["name"], tagtype["datatype"])
        else:
            table = Table("Name", "Datatype")
            for tagtype in data:
                table.add_row(tagtype["name"], tagtype["datatype"])
        console = Console()
        console.print(table)

    except:
        print("Failed to fetch tagtypes")
        raise Exception("Failed to fetch tagtypes")


@tag.command("delete")
def delete_tag(
    taguuid: Annotated[str, typer.Argument()],
):
    """
    Delete a tag
    """

    try:
        client = AuthenticatedClient()
        response = client.delete("/tag/deleteTag", params={"uuid": taguuid})
        if response.status_code < 400:
            print("Deleted tag")
        else:
            print(response)
            print("Failed to delete tag")
            raise Exception("Failed to delete tag")
    except httpx.HTTPError as e:
        print("Failed to delete tag")
        raise Exception("Failed to delete tag")



from typing import Annotated

import httpx
import typer
from rich.table import Table

from kleinkram.api_client import AuthenticatedClient

topic = typer.Typer(
    name="topic",
    help="Topic operations",
    no_args_is_help=True,
    context_settings={"help_option_names": ["-h", "--help"]},
)


@topic.command("list")
def topics(
    file: Annotated[str, typer.Option(help="Name of File")],
    full: Annotated[
        bool, typer.Option(help="As a table with additional parameters")
    ] = False,
    # Todo add mission / project as optional argument as filenames are not unique or handle multiple files
):
    """
    List topics for a file

    Only makes sense with MCAP files as we don't associate topics with BAGs as that would be redundant.
    """
    if file.endswith(".bag"):
        print("BAG files generally do not have topics")
    try:
        url = "/file/byName"
        client = AuthenticatedClient()
        response = client.get(url, params={"name": file})
        response.raise_for_status()
        data = response.json()
        if not full:
            for topic in data["topics"]:
                print(f" - {topic['name']}")
        else:
            table = Table("UUID", "name", "type", "nrMessages", "frequency")
            for topic in data["topics"]:
                table.add_row(
                    topic["uuid"],
                    topic["name"],
                    topic["type"],
                    topic["nrMessages"],
                    f"{topic['frequency']}",
                )
            print(table)

    except httpx.HTTPError as e:
        print(f"Failed")
        raise e



import glob
import os
import queue
import threading
from functools import partial
from typing import Dict
import boto3

import tqdm
from boto3.s3.transfer import TransferConfig
from botocore.client import BaseClient
from rich import print

from kleinkram.api_client import AuthenticatedClient


class TransferCallback:
    """
    Handle callbacks from the transfer manager.

    The transfer manager periodically calls the __call__ method throughout
    the upload process so that it can take action, such as displaying progress
    to the user and collecting data about the transfer.
    """

    def __init__(self):
        """
        Initialize the TransferCallback.

        This initializes an empty dictionary to hold progress bars for each file.
        """
        self._lock = threading.Lock()
        self.file_progress = {}

    def add_file(self, file_id, target_size):
        """
        Add a new file to track.

        :param file_id: A unique identifier for the file (e.g., file name or ID).
        :param target_size: The total size of the file being transferred.
        """
        with self._lock:
            tqdm_instance = tqdm.tqdm(
                total=target_size,
                unit="B",
                unit_scale=True,
                desc=f"Uploading {file_id}",
            )
            self.file_progress[file_id] = {
                "tqdm": tqdm_instance,
                "total_transferred": 0,
            }

    def __call__(self, file_id, bytes_transferred):
        """
        The callback method that is called by the transfer manager.

        Display progress during file transfer and collect per-thread transfer
        data. This method can be called by multiple threads, so shared instance
        data is protected by a thread lock.

        :param file_id: The identifier of the file being transferred.
        :param bytes_transferred: The number of bytes transferred in this call.
        """
        with self._lock:
            if file_id in self.file_progress:
                progress = self.file_progress[file_id]
                progress["total_transferred"] += bytes_transferred

                # Update tqdm progress bar
                progress["tqdm"].update(bytes_transferred)

    def close(self):
        """Close all tqdm progress bars."""
        with self._lock:
            for progress in self.file_progress.values():
                progress["tqdm"].close()


def create_transfer_callback(callback_instance, file_id):
    """
    Factory function to create a partial function for TransferCallback.
    :param callback_instance: Instance of TransferCallback.
    :param file_id: The unique identifier for the file.
    :return: A callable that can be passed as a callback to boto3's upload_file method.
    """
    return partial(callback_instance.__call__, file_id)


def expand_and_match(path_pattern):
    expanded_path = os.path.expanduser(path_pattern)
    expanded_path = os.path.expandvars(expanded_path)

    normalized_path = os.path.normpath(expanded_path)

    if "**" in normalized_path:
        file_list = glob.glob(normalized_path, recursive=True)
    else:
        file_list = glob.glob(normalized_path)

    return file_list


def uploadFiles(files: Dict[str, str], credentials: Dict[str, str], nrThreads: int):
    client = AuthenticatedClient()

    session = boto3.Session(
        aws_access_key_id=credentials["accessKey"],
        aws_secret_access_key=credentials["secretKey"],
        aws_session_token=credentials["sessionToken"],
    )
    api_endpoint = client.tokenfile.endpoint
    if api_endpoint == "http://localhost:3000":
        minio_endpoint = "http://localhost:9000"
    else:
        minio_endpoint = api_endpoint.replace("api", "minio")
    s3 = session.resource("s3", endpoint_url=minio_endpoint)

    _queue = queue.Queue()
    for file in files.items():
        _queue.put(file)
    threads = []
    transferCallback = TransferCallback()

    for i in range(nrThreads):
        thread = threading.Thread(
            target=uploadFile, args=(_queue, s3, transferCallback)
        )
        thread.start()
        threads.append(thread)
    for thread in threads:
        thread.join()


def uploadFile(_queue: queue.Queue, s3: BaseClient, transferCallback: TransferCallback):
    while True:
        try:
            filename, _file = _queue.get(timeout=3)
            queueUUID = _file["queueUUID"]
            filepath = _file["filepath"]
            bucket = _file["bucket"]
            target_location = _file["location"]
            config = TransferConfig(
                multipart_chunksize=10 * 1024 * 1024, max_concurrency=5
            )
            with open(filepath, "rb") as f:
                size = os.path.getsize(filepath)
                transferCallback.add_file(filename, size)
                callback_function = create_transfer_callback(transferCallback, filename)
                s3.Bucket(bucket).upload_file(
                    filepath, target_location, Config=config, Callback=callback_function
                )

                client = AuthenticatedClient()
                res = client.post("/queue/confirmUpload", json={"uuid": queueUUID})
                res.raise_for_status()
            _queue.task_done()
        except queue.Empty:
            break
        except Exception as e:
            print(f"Error uploading {filename}: {e}")
            _queue.task_done()


if __name__ == "__main__":
    res = expand_and_match(
        "~/Downloads/dodo_mission_2024_02_08-20240408T074313Z-003/**.bag"
    )
    print(res)



import os
from datetime import datetime, timedelta
from enum import Enum

import httpx
import pkg_resources
import typer
from rich import print
from rich.table import Table
from typer.core import TyperGroup
from typer.models import Context
from typing_extensions import Annotated

from kleinkram.api_client import AuthenticatedClient
from kleinkram.auth.auth import login, setCliKey, logout
from kleinkram.endpoint.endpoint import endpoint
from kleinkram.error_handling import ErrorHandledTyper, AccessDeniedException
from kleinkram.file.file import file
from kleinkram.mission.mission import mission
from kleinkram.project.project import project
from kleinkram.queue.queue import queue
from kleinkram.tag.tag import tag
from kleinkram.topic.topic import topic
from kleinkram.user.user import user
from .helper import uploadFiles, expand_and_match


class CommandPanel(str, Enum):
    CoreCommands = "CORE COMMANDS"
    Commands = "COMMANDS"
    AdditionalCommands = "ADDITIONAL COMMANDS"


def version_callback(value: bool):
    if value:
        typer.echo(
            f"CLI Version: {pkg_resources.get_distribution('kleinkram').version}"
        )
        raise typer.Exit()


class OrderCommands(TyperGroup):
    """

    The following code snippet is taken from https://github.com/tiangolo/typer/discussions/855 (see comment
    https://github.com/tiangolo/typer/discussions/855#discussioncomment-9824582) and adapted to our use case.
    """

    def list_commands(self, _ctx: Context) -> list[str]:
        order = list(CommandPanel)
        grouped_commands = {
            name: getattr(command, "rich_help_panel")
            for name, command in sorted(self.commands.items())
            if getattr(command, "rich_help_panel") in order
        }
        ungrouped_command_names = [
            command.name
            for command in self.commands.values()
            if command.name not in grouped_commands
        ]
        return [
            name
            for name, command in sorted(
                grouped_commands.items(),
                key=lambda item: order.index(item[1]),
            )
        ] + sorted(ungrouped_command_names)


app = ErrorHandledTyper(
    context_settings={"help_option_names": ["-h", "--help"]},
    no_args_is_help=True,
    cls=OrderCommands,
)


@app.callback()
def version(
    version: bool = typer.Option(
        None,
        "--version",
        "-v",
        callback=version_callback,
        is_eager=True,
        help="Print the version and exit",
    )
):
    pass


app.add_typer(project, rich_help_panel=CommandPanel.Commands)
app.add_typer(mission, rich_help_panel=CommandPanel.Commands)

app.add_typer(topic, rich_help_panel=CommandPanel.Commands)
app.add_typer(file, rich_help_panel=CommandPanel.Commands)
app.add_typer(queue, rich_help_panel=CommandPanel.Commands)
app.add_typer(user, rich_help_panel=CommandPanel.Commands)
app.add_typer(tag, rich_help_panel=CommandPanel.Commands)
app.add_typer(endpoint, rich_help_panel=CommandPanel.AdditionalCommands)

app.command(rich_help_panel=CommandPanel.AdditionalCommands)(login)
app.command(rich_help_panel=CommandPanel.AdditionalCommands)(logout)
app.command(hidden=True)(setCliKey)


@app.command("download", rich_help_panel=CommandPanel.CoreCommands)
def download():
    raise NotImplementedError("Not implemented yet.")


@app.command("upload", rich_help_panel=CommandPanel.CoreCommands)
def upload(
    path: Annotated[
        str, typer.Option(prompt=True, help="Path to files to upload, Regex supported")
    ],
    project: Annotated[str, typer.Option(prompt=True, help="Name of Project")],
    mission: Annotated[
        str, typer.Option(prompt=True, help="Name of Mission to create")
    ],
):
    """
    Upload files matching the path to a mission in a project.

    The mission name must be unique within the project and not yet created.\n
    Examples:\n
        - 'klein upload --path "~/data/**/*.bag" --project "Project 1" --mission "Mission 1"'\n

    """
    files = expand_and_match(path)
    filenames = list(
        map(lambda x: x.split("/")[-1], filter(lambda x: not os.path.isdir(x), files))
    )
    if not filenames:
        raise ValueError("No files found matching the given path.")

    print(
        f"Uploading the following files to mission '{mission}' in project '{project}':"
    )
    filepaths = {}
    for path in files:
        if not os.path.isdir(path):
            filepaths[path.split("/")[-1]] = path
            typer.secho(f" - {path}", fg=typer.colors.RESET)

    try:
        client = AuthenticatedClient()

        get_project_url = "/project/byName"
        project_response = client.get(get_project_url, params={"name": project})
        if project_response.status_code >= 400:

            raise AccessDeniedException(
                f"The project '{project}' does not exist or you do not have access to it.\n"
                f"Consider using the following command to create a project: 'klein project create'\n",
                f"{project_response.json()['message']} ({project_response.status_code})",
            )

        project_json = project_response.json()
        if not project_json["uuid"]:
            print(f"Project not found: '{project}'")
            return

        get_mission_url = "/mission/byName"
        mission_response = client.get(get_mission_url, params={"name": mission})
        mission_response.raise_for_status()
        if mission_response.content:
            mission_json = mission_response.json()
            if mission_json["uuid"]:
                raise ValueError(
                    f"Mission {mission_json['name']} ({mission_json['uuid']}) already exists. Delete it or select "
                    f"another name."
                )
            raise Exception(f"Something failed, should not happen")

        create_mission_url = "/mission/create"
        new_mission = client.post(
            create_mission_url,
            json={"name": mission, "projectUUID": project_json["uuid"], "tags": []},
        )
        new_mission.raise_for_status()
        new_mission_data = new_mission.json()

        get_temporary_credentials = "/file/temporaryAccess"

        response_2 = client.post(
            get_temporary_credentials,
            json={"filenames": filenames, "missionUUID": new_mission_data["uuid"]},
        )
        response_2.raise_for_status()
        temp_credentials = response_2.json()
        credential = temp_credentials["credentials"]
        confirmed_files = temp_credentials["files"]
        for _file in filenames:
            if not _file in confirmed_files.keys():
                raise Exception(
                    "Could not upload File '" + file + "'. Is the filename unique? "
                )
            confirmed_files[_file]["filepath"] = filepaths[_file]
        if len(confirmed_files.keys()) > 0:
            uploadFiles(confirmed_files, credential, 4)

    except httpx.HTTPError as e:
        print(e)


@queue.command("list")
def list_queue():
    """List current Queue entities"""
    try:
        url = "/queue/active"
        startDate = datetime.now().date() - timedelta(days=1)
        client = AuthenticatedClient()
        response = client.get(url, params={"startDate": startDate})
        response.raise_for_status()
        data = response.json()
        table = Table("UUID", "filename", "mission", "state", "origin", "createdAt")
        for topic in data:
            table.add_row(
                topic["uuid"],
                topic["filename"],
                topic["mission"]["name"],
                topic["state"],
                topic["location"],
                topic["createdAt"],
            )
        print(table)

    except httpx.HTTPError as e:
        print(e)


@app.command("claim", hidden=True)
def claim():
    """
    Claim admin rights as the first user

    Only works if no other user has claimed admin rights before.
    """

    client = AuthenticatedClient()
    response = client.post("/user/claimAdmin")
    response.raise_for_status()
    print("Admin claimed.")


if __name__ == "__main__":
    app()
